import urllib.request as urlreq
import threading
import json

from .api import api as API
from .handler import Handler
from .config import config
from .events import event_types

class Worker(threading.Thread):
    def __init__(self, func, args = None):
        threading.Thread.__init__(self)
        self.func = func
        self.args = args
    def run(self):
        try:
            self.func(self.args)
        except Exception as er:
            return 0

class Longpoll:
    def __init__(self, token = None, v = "5.95", api = API):
        self.token = token
        self.api_v = v
        if self.token is None:
            raise RuntimeError("You must past token into settings")

        config['access_token'] = self.token
        config['api_version'] = self.api_v

        self.api = api

        self.group = self.api("groups.getById")[0]
        self.lp_server_data = None
        self.lp_server = "{}?act=a_check&key={}&ts={}&wait=25"

        self.is_polling = None

        # other
        self.handler = Handler()
        self.command = self.handler.command
        self.event = self.handler.event


    def get_longpoll_data(self):
        longpoll = self.api("groups.getLongPollServer", group_id = self.group['id'])

        self.lp_server_data = {
            **longpoll
        }

    def receiver(self):
        if not self.lp_server_data:
            self.get_longpoll_data()

        try:
            with urlreq.urlopen(
                self.lp_server.format(
                    self.lp_server_data['server'],
                    self.lp_server_data['key'],
                    self.lp_server_data['ts']
                )
            ) as res:
                response = json.loads(res.read())
        except Exception:
            return ()

        if 'error' in response:
            self.get_longpoll_data()

        if 'ts' in response:
            self.lp_server_data['ts'] = response['ts']

        return response['updates']

    def main(self, none):
        try:
            while True:
                if not self.is_polling:
                    break
                for update in self.receiver():
                    if update['type'] in event_types:
                        worker = Worker(self.handler._update_handler(update), self.handler._msg_parser(update['object']))
                        worker.start()
        except KeyboardInterrupt:
            self.is_polling = False

    def startPolling(self):
        self.is_polling = True
        worker = Worker(self.main)
        worker.start()

    def stopPolling(self):
        self.is_polling = False




from .utils import Utils
from objdict import ObjDict

class Handler:
    def __init__(self):
        self._msg_command_handler_list = []
        self._event_handler_list = []

    def _msg_parser(self, obj):
        object = dict(object = obj)
        utils = Utils(obj)

        available_methods = [m for m in dir(utils) if callable(getattr(utils, m)) and not m.startswith("_")]

        for method in available_methods:
            object[method] = getattr(utils, method)

        return ObjDict(object)

    def _msg_command_handler(self, object):
        if not self._msg_command_handler_list:
            return ()
            
        for update in self._msg_command_handler_list:
            command = object['text'].split(" ").pop(0)
            if command in update['commands']:
                return update['object']

    def _event_handler(self, object):
        for update in self._event_handler_list:
            if object['type'] in update['events']:
                return update['object']

    def _update_handler(self, update):
        if update['type'] == "message_new" and self._msg_command_handler_list:
            return self._msg_command_handler(update['object'])
        else:
            return self._event_handler(update)
        
    def command(self, commands = None):
        def decorator(object):
            def wrapper():
                self._msg_command_handler_list.append(
                    dict(
                        commands = commands,
                        object = object
                    )
                )
                
                return object
            return wrapper()
        return decorator

    def event(self, events = None):
        def decorator(object):
            def wrapper():
                self._event_handler_list.append(
                    dict(
                        events = events,
                        object = object
                    )
                )

                return object
            return wrapper()
        return decorator


import urllib.request as urlreq
import urllib.parse as urlparse
import json

from .config import config

def api(method, **kwargs):
    api_url = "https://api.vk.com/method/{{}}?access_token={}&v={}".format(config['access_token'], config['api_version'])

    method_url = api_url.format(method)
    data = {}
    for k, v in kwargs.items():
        if v is not None:
            data[k] = v
    try:
        with urlreq.urlopen(
            url = method_url, 
            data = urlparse.urlencode(data).encode('utf-8')
        ) as r:
            res = json.loads(r.read().decode('utf-8'))
    except Exception as er:
        raise RuntimeError(er)

    if 'error' in res:
        return res['error']

    return res['response']



from .api import api
from .keyboard import Keyboard

class Utils:
    def __init__(self, data):
        self.data = data
        self._kb = Keyboard()

    def sendMessage(self, peer_id, message, attachment = None, sticker_id = None, keyboard = None, **kwargs):
        return api(
            "messages.send",
            message = str(message),
            random_id = 0,
            peer_id = int(peer_id),
            attachment = attachment,
            sticker_id = sticker_id,
            keyboard = keyboard,
            **kwargs
        )

    def reply(self, message, **kwargs):
        return self.sendMessage(
            message = str(message),
            peer_id = int(self.data['peer_id']),
            **kwargs
        )
    
    def keyboard(self, buttons, one_time = False):
        return self._kb(
            buttons = buttons,
            one_time = one_time
        )




import json

class Keyboard:
    def __init__(self):
        self.colors = ('default', 'primary', 'positive', 'negative')
        self.one_time = None
        self.buttons = []

    def __call__(self, buttons, one_time = False):
        self.one_time = one_time
        return self._check_buttons(buttons)
    
    def _check_buttons(self, buttons):
        if len(buttons) > 10:
            raise ValueError("The number of rows exceeds the limit (max rows: 10)")

        for row_num, row in enumerate(buttons, 1):
            if len(row) > 4:
                raise ValueError(f"The number of buttons in the same row exceeds the limit (row: {row_num})")

            for btn_num, btn in enumerate(row, 1):
                if 'label' not in btn:
                    raise ValueError(f"The field must not be empty: 'label' (row: {row_num}, button: {btn_num})") 
                if 'color' not in btn:
                    raise ValueError(f"The field must not be empty: 'color' (row: {row_num}, button: {btn_num})") 

                if btn['color'] not in self.colors:
                    raise ValueError("Wrong button color! Available colors: {}".format(", ".join(self.colors)))

        return self._dump_keyboard(buttons)

    def _dump_keyboard(self, buttons):
        btn_num = 1
        for row in buttons:
            formatted_row = []
            for btn in row:
                btn_template = dict(
                    action = dict(
                        type = "text",
                        payload = "{\"button\": "f"{btn_num}""}",
                        label = btn['label']
                    ),
                    color = btn['color']
                )
                formatted_row.append(btn_template)
                btn_num += 1
                print(btn_num)
            self.buttons.append(formatted_row)

        return json.dumps(dict(
            one_time = self.one_time,
            buttons = self.buttons,
        ), ensure_ascii = False)



event_types = [
    "message_new",
    "message_reply",
    "message_edit",
    "message_allow",
    "message_deny",
    "photo_new",
    "photo_comment_new",
    "photo_comment_edit",
    "photo_comment_restore",
    "photo_comment_delete",
    "audio_new",
    "video_new",
    "video_comment_new",
    "video_comment_edit",
    "video_comment_restore",
    "video_comment_delete",
    "wall_post_new",
    "wall_repost",
    "wall_reply_new",
    "wall_reply_edit",
    "wall_reply_restore",
    "wall_reply_delete",
    "board_post_new",
    "board_post_edit",
    "board_post_restore",
    "board_post_delete",
    "market_comment_new",
    "market_comment_edit",
    "market_comment_restore",
    "market_comment_delete",
    "group_leave",
    "group_join",
    "user_block",
    "user_unblock",
    "poll_vote_new",
    "group_officers_edit",
    "group_change_settings",
    "group_change_photo",
    "vkpay_transaction"
]